from typing import Any, Dict, Iterable, List, Optional, Set, Union

from langchain_core._api import beta
from langchain_core.documents import Document

from langchain_community.graph_vectorstores.extractors.link_extractor import (
    LinkExtractor,
)
from langchain_community.graph_vectorstores.links import Link

# TypeAlias is not available in Python 3.9, we can't use that or the newer `type`.
GLiNERInput = Union[str, Document]


@beta()
class GLiNERLinkExtractor(LinkExtractor[GLiNERInput]):
    """Link documents with common named entities using `GLiNER`_.

    `GLiNER`_ is a Named Entity Recognition (NER) model capable of identifying any
    entity type using a bidirectional transformer encoder (BERT-like).

    The ``GLiNERLinkExtractor`` uses GLiNER to create links between documents that
    have named entities in common.

    Example::

        extractor = GLiNERLinkExtractor(
            labels=["Person", "Award", "Date", "Competitions", "Teams"]
        )
        results = extractor.extract_one("some long text...")

    .. _GLiNER: https://github.com/urchade/GLiNER

    .. seealso::

            - :mod:`How to use a graph vector store <langchain_community.graph_vectorstores>`
            - :class:`How to create links between documents <langchain_community.graph_vectorstores.links.Link>`

    How to link Documents on common named entities
    ==============================================

    Preliminaries
    -------------

    Install the ``gliner`` package:

    .. code-block:: bash

        pip install -q langchain_community gliner

    Usage
    -----

    We load the ``state_of_the_union.txt`` file, chunk it, then for each chunk we
    extract named entity links and add them to the chunk.

    Using extract_one()
    ^^^^^^^^^^^^^^^^^^^

    We can use :meth:`extract_one` on a document to get the links and add the links
    to the document metadata with
    :meth:`~langchain_community.graph_vectorstores.links.add_links`::

        from langchain_community.document_loaders import TextLoader
        from langchain_community.graph_vectorstores import CassandraGraphVectorStore
        from langchain_community.graph_vectorstores.extractors import GLiNERLinkExtractor
        from langchain_community.graph_vectorstores.links import add_links
        from langchain_text_splitters import CharacterTextSplitter

        loader = TextLoader("state_of_the_union.txt")
        raw_documents = loader.load()

        text_splitter = CharacterTextSplitter(chunk_size=1000, chunk_overlap=0)
        documents = text_splitter.split_documents(raw_documents)

        ner_extractor = GLiNERLinkExtractor(["Person", "Topic"])
        for document in documents:
            links = ner_extractor.extract_one(document)
            add_links(document, links)

        print(documents[0].metadata)

    .. code-block:: output

        {'source': 'state_of_the_union.txt', 'links': [Link(kind='entity:Person', direction='bidir', tag='President Zelenskyy'), Link(kind='entity:Person', direction='bidir', tag='Vladimir Putin')]}

    Using LinkExtractorTransformer
    ^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

    Using the :class:`~langchain_community.graph_vectorstores.extractors.link_extractor_transformer.LinkExtractorTransformer`,
    we can simplify the link extraction::

        from langchain_community.document_loaders import TextLoader
        from langchain_community.graph_vectorstores.extractors import (
            GLiNERLinkExtractor,
            LinkExtractorTransformer,
        )
        from langchain_text_splitters import CharacterTextSplitter

        loader = TextLoader("state_of_the_union.txt")
        raw_documents = loader.load()

        text_splitter = CharacterTextSplitter(chunk_size=1000, chunk_overlap=0)
        documents = text_splitter.split_documents(raw_documents)

        ner_extractor = GLiNERLinkExtractor(["Person", "Topic"])
        transformer = LinkExtractorTransformer([ner_extractor])
        documents = transformer.transform_documents(documents)

        print(documents[0].metadata)

    .. code-block:: output

        {'source': 'state_of_the_union.txt', 'links': [Link(kind='entity:Person', direction='bidir', tag='President Zelenskyy'), Link(kind='entity:Person', direction='bidir', tag='Vladimir Putin')]}

    The documents with named entity links can then be added to a :class:`~langchain_community.graph_vectorstores.base.GraphVectorStore`::

        from langchain_community.graph_vectorstores import CassandraGraphVectorStore

        store = CassandraGraphVectorStore.from_documents(documents=documents, embedding=...)

    Args:
        labels: List of kinds of entities to extract.
        kind: Kind of links to produce with this extractor.
        model: GLiNER model to use.
        extract_kwargs: Keyword arguments to pass to GLiNER.
    """  # noqa: E501

    def __init__(
        self,
        labels: List[str],
        *,
        kind: str = "entity",
        model: str = "urchade/gliner_mediumv2.1",
        extract_kwargs: Optional[Dict[str, Any]] = None,
    ):
        try:
            from gliner import GLiNER

            self._model = GLiNER.from_pretrained(model)

        except ImportError:
            raise ImportError(
                "gliner is required for GLiNERLinkExtractor. "
                "Please install it with `pip install gliner`."
            ) from None

        self._labels = labels
        self._kind = kind
        self._extract_kwargs = extract_kwargs or {}

    def extract_one(self, input: GLiNERInput) -> Set[Link]:  # noqa: A002
        return next(iter(self.extract_many([input])))

    def extract_many(
        self,
        inputs: Iterable[GLiNERInput],
    ) -> Iterable[Set[Link]]:
        strs = [i if isinstance(i, str) else i.page_content for i in inputs]
        for entities in self._model.batch_predict_entities(
            strs, self._labels, **self._extract_kwargs
        ):
            yield {
                Link.bidir(kind=f"{self._kind}:{e['label']}", tag=e["text"])
                for e in entities
            }



from collections.abc import Iterable
from dataclasses import dataclass
from typing import Literal, Union

from langchain_core._api import beta
from langchain_core.documents import Document


@beta()
@dataclass(frozen=True)
class Link:
    """A link to/from a tag of a given kind.

    Documents in a :class:`graph vector store <langchain_community.graph_vectorstores.base.GraphVectorStore>`
    are connected via "links".
    Links form a bipartite graph between documents and tags: documents are connected
    to tags, and tags are connected to other documents.
    When documents are retrieved from a graph vector store, a pair of documents are
    connected with a depth of one if both documents are connected to the same tag.

    Links have a ``kind`` property, used to namespace different tag identifiers.
    For example a link to a keyword might use kind ``kw``, while a link to a URL might
    use kind ``url``.
    This allows the same tag value to be used in different contexts without causing
    name collisions.

    Links are directed. The directionality of links controls how the graph is
    traversed at retrieval time.
    For example, given documents ``A`` and ``B``, connected by links to tag ``T``:

    +----------+----------+---------------------------------+
    | A to T   | B to T   | Result                          |
    +==========+==========+=================================+
    | outgoing | incoming | Retrieval traverses from A to B |
    +----------+----------+---------------------------------+
    | incoming | incoming | No traversal from A to B        |
    +----------+----------+---------------------------------+
    | outgoing | incoming | No traversal from A to B        |
    +----------+----------+---------------------------------+
    | bidir    | incoming | Retrieval traverses from A to B |
    +----------+----------+---------------------------------+
    | bidir    | outgoing | No traversal from A to B        |
    +----------+----------+---------------------------------+
    | outgoing | bidir    | Retrieval traverses from A to B |
    +----------+----------+---------------------------------+
    | incoming | bidir    | No traversal from A to B        |
    +----------+----------+---------------------------------+

    Directed links make it possible to describe relationships such as term
    references / definitions: term definitions are generally relevant to any documents
    that use the term, but the full set of documents using a term generally aren't
    relevant to the term's definition.

    .. seealso::

        - :mod:`How to use a graph vector store <langchain_community.graph_vectorstores>`
        - :class:`How to link Documents on hyperlinks in HTML <langchain_community.graph_vectorstores.extractors.html_link_extractor.HtmlLinkExtractor>`
        - :class:`How to link Documents on common keywords (using KeyBERT) <langchain_community.graph_vectorstores.extractors.keybert_link_extractor.KeybertLinkExtractor>`
        - :class:`How to link Documents on common named entities (using GliNER) <langchain_community.graph_vectorstores.extractors.gliner_link_extractor.GLiNERLinkExtractor>`

    How to add links to a Document
    ==============================

    How to create links
    -------------------

    You can create links using the Link class's constructors :meth:`incoming`,
    :meth:`outgoing`, and :meth:`bidir`::

        from langchain_community.graph_vectorstores.links import Link

        print(Link.bidir(kind="location", tag="Paris"))

    .. code-block:: output

        Link(kind='location', direction='bidir', tag='Paris')

    Extending documents with links
    ------------------------------

    Now that we know how to create links, let's associate them with some documents.
    These edges will strengthen the connection between documents that share a keyword
    when using a graph vector store to retrieve documents.

    First, we'll load some text and chunk it into smaller pieces.
    Then we'll add a link to each document to link them all together::

        from langchain_community.document_loaders import TextLoader
        from langchain_community.graph_vectorstores.links import add_links
        from langchain_text_splitters import CharacterTextSplitter

        loader = TextLoader("state_of_the_union.txt")

        raw_documents = loader.load()
        text_splitter = CharacterTextSplitter(chunk_size=1000, chunk_overlap=0)
        documents = text_splitter.split_documents(raw_documents)

        for doc in documents:
            add_links(doc, Link.bidir(kind="genre", tag="oratory"))

        print(documents[0].metadata)

    .. code-block:: output

        {'source': 'state_of_the_union.txt', 'links': [Link(kind='genre', direction='bidir', tag='oratory')]}

    As we can see, each document's metadata now includes a bidirectional link to the
    genre ``oratory``.

    The documents can then be added to a graph vector store::

        from langchain_community.graph_vectorstores import CassandraGraphVectorStore

        graph_vectorstore = CassandraGraphVectorStore.from_documents(
            documents=documents, embeddings=...
        )

    """  # noqa: E501

    kind: str
    """The kind of link. Allows different extractors to use the same tag name without
    creating collisions between extractors. For example “keyword” vs “url”."""
    direction: Literal["in", "out", "bidir"]
    """The direction of the link."""
    tag: str
    """The tag of the link."""

    @staticmethod
    def incoming(kind: str, tag: str) -> "Link":
        """Create an incoming link.

        Args:
            kind: the link kind.
            tag: the link tag.
        """
        return Link(kind=kind, direction="in", tag=tag)

    @staticmethod
    def outgoing(kind: str, tag: str) -> "Link":
        """Create an outgoing link.

        Args:
            kind: the link kind.
            tag: the link tag.
        """
        return Link(kind=kind, direction="out", tag=tag)

    @staticmethod
    def bidir(kind: str, tag: str) -> "Link":
        """Create a bidirectional link.

        Args:
            kind: the link kind.
            tag: the link tag.
        """
        return Link(kind=kind, direction="bidir", tag=tag)


METADATA_LINKS_KEY = "links"


@beta()
def get_links(doc: Document) -> list[Link]:
    """Get the links from a document.

    Args:
        doc: The document to get the link tags from.
    Returns:
        The set of link tags from the document.
    """

    links = doc.metadata.setdefault(METADATA_LINKS_KEY, [])
    if not isinstance(links, list):
        # Convert to a list and remember that.
        links = list(links)
        doc.metadata[METADATA_LINKS_KEY] = links
    return links


@beta()
def add_links(doc: Document, *links: Union[Link, Iterable[Link]]) -> None:
    """Add links to the given metadata.

    Args:
        doc: The document to add the links to.
        *links: The links to add to the document.
    """
    links_in_metadata = get_links(doc)
    for link in links:
        if isinstance(link, Iterable):
            links_in_metadata.extend(link)
        else:
            links_in_metadata.append(link)


@beta()
def copy_with_links(doc: Document, *links: Union[Link, Iterable[Link]]) -> Document:
    """Return a document with the given links added.

    Args:
        doc: The document to add the links to.
        *links: The links to add to the document.

    Returns:
        A document with a shallow-copy of the metadata with the links added.
    """
    new_links = set(get_links(doc))
    for link in links:
        if isinstance(link, Iterable):
            new_links.update(link)
        else:
            new_links.add(link)

    return Document(
        page_content=doc.page_content,
        metadata={
            **doc.metadata,
            METADATA_LINKS_KEY: list(new_links),
        },
    )



"""Tool for the SearxNG search API."""

from typing import Optional, Type

from langchain_core.callbacks import (
    AsyncCallbackManagerForToolRun,
    CallbackManagerForToolRun,
)
from langchain_core.tools import BaseTool
from pydantic import BaseModel, ConfigDict, Field

from langchain_community.utilities.searx_search import SearxSearchWrapper


class SearxSearchQueryInput(BaseModel):
    """Input for the SearxSearch tool."""

    query: str = Field(description="query to look up on searx")


class SearxSearchRun(BaseTool):
    """Tool that queries a Searx instance."""

    name: str = "searx_search"
    description: str = (
        "A meta search engine."
        "Useful for when you need to answer questions about current events."
        "Input should be a search query."
    )
    wrapper: SearxSearchWrapper
    kwargs: dict = Field(default_factory=dict)
    args_schema: Type[BaseModel] = SearxSearchQueryInput

    def _run(
        self,
        query: str,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> str:
        """Use the tool."""
        return self.wrapper.run(query, **self.kwargs)

    async def _arun(
        self,
        query: str,
        run_manager: Optional[AsyncCallbackManagerForToolRun] = None,
    ) -> str:
        """Use the tool asynchronously."""
        return await self.wrapper.arun(query, **self.kwargs)


class SearxSearchResults(BaseTool):
    """Tool that queries a Searx instance and gets back json."""

    name: str = "searx_search_results"
    description: str = (
        "A meta search engine."
        "Useful for when you need to answer questions about current events."
        "Input should be a search query. Output is a JSON array of the query results"
    )
    wrapper: SearxSearchWrapper
    num_results: int = 4
    kwargs: dict = Field(default_factory=dict)
    args_schema: Type[BaseModel] = SearxSearchQueryInput

    model_config = ConfigDict(
        extra="allow",
    )

    def _run(
        self,
        query: str,
        run_manager: Optional[CallbackManagerForToolRun] = None,
    ) -> str:
        """Use the tool."""
        return str(self.wrapper.results(query, self.num_results, **self.kwargs))

    async def _arun(
        self,
        query: str,
        run_manager: Optional[AsyncCallbackManagerForToolRun] = None,
    ) -> str:
        """Use the tool asynchronously."""
        return (
            await self.wrapper.aresults(query, self.num_results, **self.kwargs)
        ).__str__()
